from typing import Tuple, Dict, List
from .papers_display import papers_results2message
from .papers_args import add_arguments_papers, parse_args_papers


def search_by_title_args2querys(args: object) -> List[Tuple[str, Dict]]:
    where, orderby, limits, values = parse_args_papers(args)
    return [(
        f"MATCH (c:Publication)-[:CITE]->(p:Publication)-[:PUBLISH]->(j:Journal) WHERE {where} "
        f"RETURN p, j, COUNT(c) AS citation ORDER BY {orderby} LIMIT {limits}",
        values
    )]


search_by_title_jump = (
    "search_by_title",
    add_arguments_papers,
    search_by_title_args2querys,
    papers_results2message,
    "Search papers by keywords in title"
)



from argparse import ArgumentParser
from enum import Enum
from datetime import datetime


class Order(Enum):
    Citation = 'citation'
    Year = 'year'
    Date = 'date'

    def __str__(self):
        return self.value


def add_arguments_papers(parser: ArgumentParser):
    parser.add_argument("-o", "--order", type=Order, choices=list(Order), default=Order.Citation,
                        help="Order by what.")
    parser.add_argument("-y", "--year", type=int, default=datetime.now().year - 5,
                        help="Donot include those papers before this year.")
    parser.add_argument("-k", "--keyword", action='append', default=[],
                        help="Specify keyword rules for searching the title. "
                        "-k 'word1 word2' means 'word1 and word2', "
                        "-k 'word1' -k 'word2' means 'word1 or word2'. "
                        "e.g. -k 'video edge' -k 'super-resolution' means search for those papers "
                        "whose title includes both 'video' and 'edge' or includes 'super-resolution'")
    parser.add_argument("-v", "--key_value", action='append', default=[],
                        help="Specify key-value rules for searching in the database. e.g. -v j.ccf:A will only return those CCF A papers")
    parser.add_argument("-l", "--limit", type=int, help="Limit the length of results. Maximum 20", default=20)
    return parser


def parse_args_papers(args: object):
    orderby = 'citation DESC'
    if args.order == Order.Year:
        orderby = 'p.year DESC, citation DESC'
    elif args.order == Order.Date:
        orderby = 'p.year DESC, COALESCE(p.date, date("1970-01-01")) DESC, citation DESC'

    limits = "$limit"
    where = 'p.year >= $year'
    values = dict(year=args.year, limit=args.limit)

    ki, k_or, v_or = 0, [], {}
    for keywords in args.keyword:
        k_and, v_and = [], {}
        for k in keywords.split(" "):
            if not k:
                continue
            ki += 1
            k_and.append(f"toLower(p.title) CONTAINS $keyword{ki}")
            v_and[f"keyword{ki}"] = k
        k_or.append(f"({' and '.join(k_and)})")
        v_or = {**v_or, **v_and}
    if ki > 0:
        where += " AND " + f"({' OR '.join(k_or)})"
        values = {**values, **v_or}
    
    ki, keys, kvalues = 0, [], {}
    for key_value in args.key_value:
        kv = key_value.split(":", 1)
        if len(kv) != 2 or not kv[0]:
            continue
        ki += 1
        key, value = kv
        keys.append(f"{key}=$value{ki}")
        kvalues[f"value{ki}"] = value
    if ki > 0:
        where += " AND " + f"({' AND '.join(keys)})"
        values = {**values, **kvalues}

    return where, orderby, limits, values



import argparse
import logging
from neo4j import GraphDatabase
from telegram.ext import Application, ApplicationBuilder
from .arg import add_argument_jump, parse_args_jump
from .rabbot import Rabbot
from .start import start_args2querys, gen_start_results2message
from .jumps import author_papers_jump, citations_jump, references_jump, search_by_title_jump, paper_detail_jump
default_jumps = [
    search_by_title_jump, author_papers_jump, citations_jump, references_jump, paper_detail_jump
]
default_jumps_name = [
    "start_jump", "author_papers_jump", "citations_jump", "references_jump", "search_by_title_jump", "paper_detail_jump"
]
defaults_desc = "".join([" -j "+name for name in default_jumps_name])
logging.basicConfig(
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
    level=logging.INFO
)

parser = argparse.ArgumentParser()

parser.add_argument("--auth", type=str, default=None, help=f'Auth to neo4j database.')
parser.add_argument("--uri", type=str, required=True, help=f'URI to neo4j database.')
parser.add_argument("--token", type=str, required=True, help=f'Telegram bot token.')
add_argument_jump(parser, defaults_desc=defaults_desc)


args = parser.parse_args()
jump_list = parse_args_jump(parser)

# Parse jump list
jump_dict = {name: (parser, message2query, result2message)
             for name, parser, message2query, result2message, _ in jump_list}
desc_dict = {name: desc for name, _, _, _, desc in jump_list if desc}
desc_order = [name for name, _, _, _, desc in jump_list if desc]
# Add default jumps
for name, parser, message2query, result2message, desc in default_jumps:
    if name not in jump_dict:
        jump_dict[name] = (parser, message2query, result2message)
    if desc:
        desc_dict[name] = desc
        desc_order.append(name)


async def post_init(application: Application):
    await application.bot.set_my_commands([('start', "Let's jump!")])
application = ApplicationBuilder().token(args.token).post_init(post_init).build()

with GraphDatabase.driver(args.uri, auth=args.auth) as driver:
    with driver.session() as session:
        rabbot = Rabbot(app=application, session=session)
        for name, (parser, message2query, result2message) in jump_dict.items():
            rabbot.add_jump(name, parser, message2query, result2message)
        rabbot.add_jump('start', None, start_args2querys, gen_start_results2message(desc_order, desc_dict))
        application.run_polling()
